#include <stdio.h>
#include <stdlib.h>
#include <mpi.h>

// maximum number of each value in matrix or vector
#define NUMMAX 10

void init_array(long *data, int len, unsigned int seed);

// row count in a matrix
#define M_ROW 200
// column count in a matrix
#define M_COL 100


int main(int argc, char *argv[])
{
  long *matrix, *vector, *answer;
  int i, j;

  long *each_mat, *each_ans;
  int rank, nprocs, nmyrow;

  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  MPI_Comm_size(MPI_COMM_WORLD, &nprocs);
  nmyrow = M_ROW / nprocs; //割り切れるように、M_ROW、nprocsを調整する

  // すべてのプロセスで共通のメモリ領域を確保
  each_mat = (long*)malloc(nmyrow * M_COL * sizeof(long));
  vector = (long*)malloc(M_COL * sizeof(long));
  each_ans = (long*)malloc(nmyrow * sizeof(long));
  if (each_mat == NULL || vector == NULL || each_ans == NULL) {
    fprintf(stderr, "malloc error\n");
    return 1;
  }

  // 乱数を用いて配列データを初期化するので、特定1プロセス上でのみ
  // データの初期化を行う。
  if (rank == 0) {
    matrix = (long*)malloc(M_ROW * M_COL * sizeof(long));
    answer = (long*)malloc(M_ROW * sizeof(long));
    if (matrix == NULL || answer == NULL) {
      fprintf(stderr, "malloc error\n");
      MPI_Abort(MPI_COMM_WORLD, 1);
    }
    init_array(matrix, M_ROW * M_COL, rank);
    init_array(vector, M_COL, rank + 1);
  }

  // データの配布
  MPI_Scatter(matrix,   nmyrow * M_COL, MPI_LONG,
              each_mat, nmyrow * M_COL, MPI_LONG,
              0, MPI_COMM_WORLD);
  MPI_Bcast(vector, M_COL, MPI_LONG, 0, MPI_COMM_WORLD);

  /* calc */
  for (i = 0; i < nmyrow; i++) {
    each_ans[i] = 0;
    for (j = 0; j < M_COL; j++) {
      each_ans[i] += each_mat[M_COL*i+j] * vector[j];
    }
  }

  // 計算結果の集計
  MPI_Gather(each_ans, nmyrow, MPI_LONG,
             answer,   nmyrow, MPI_LONG,
             0, MPI_COMM_WORLD);

  /* print */
  if (rank == 0) {
    printf("Matrix\n");
    for (i = 0; i < M_ROW; i++) {
      printf("  ");
      for (j = 0; j < M_COL; j++) {
        printf("%ld ", matrix[M_COL*i+j]);
      }
      printf("\n");
    }
    printf("Vector\n");
    for (i = 0; i < M_COL; i++)
      printf("  %ld\n", vector[i]);
    printf("\n");
    printf("Answer\n");
    for (i = 0; i < M_COL; i++)
      printf("  %ld\n", answer[i]);

    free(matrix);
    free(answer);
  }

  free(each_mat);
  free(vector);
  free(each_ans);

  MPI_Finalize();
  return 0;
}


void init_array(long *data, int len, unsigned int seed)
{
  int i;
  srandom(seed);
  for (i = 0; i < len; i++) {
    data[i] = random() % NUMMAX;
    if (random() % 2 == 0) data[i] = -1 * data[i];
  }
}
